#ifndef GRAPH_H
#define GRAPH_H

#include <qwidget.h>
#include <qobject.h>
#include <qarray.h>
#include <qfont.h>
#include <qpen.h>

#include <qwt_plot.h>
#include <qwt_marker.h>
#include <qwt_plot_classes.h>

#include "curvesDialog.h"
#include "plotDialog.h"
#include "worksheet.h"
#include "genDialog.h"

class QPopupMenu;
class QStringList;
class QTabDialog;
class QRadioButton;
class QLineEdit;
class QComboBox;
class QMainWindow;
class QSpinBox;
class QwtPlotZoomer;
	
class WksheetWindow;
class LegendMarker;
class LineMarker;
class ImageMarker;
class TitlePicker;
class ScalePicker;
class CanvasPicker;
class QwtPieCurve;
class Plot;
	
class Graph: public QWidget
{
	Q_OBJECT

public:
    Graph (QWidget* parent=0, const char* name=0, WFlags f=0);
	~Graph();

	enum AxisType{Numeric = 0, Txt = 1, Day = 2, Month = 3};

	enum MarkerType{None=-1, Text = 0, Arrow=1, Image=2};

	enum LabelFormat{Automatic = 0, Decimal = 1, Scientific = 2, Superscripts = 3};
	
	enum PlotType{Line = 0,
				  Scatter = 1, 
				  LineSymbols = 2,
				  VerticalBars = 3,
				  Area = 4,
				  Pie = 5,
				  VerticalDropLines = 6,
				  Spline = 7,
				  Steps = 8,
				  Histogram = 9,
				  HorizontalBars = 10,
				  VectXYXY = 11};

	Plot *myPlot;
	QwtPlotZoomer *d_zoomer;
	TitlePicker *titlePicker;
	ScalePicker *scalePicker;
	CanvasPicker* cp;
	 
public slots:
	QwtPlot* plotWidget();

	// pie curves	
	bool isPiePlot();
	QwtPieCurve* pieCurve();
	void plotPie(QwtPieCurve* curve);
	void plotPie(WksheetWindow* w,const QString& name);
	void plotPie(WksheetWindow* w,const QString& name,const QPen& pen, int brush, int size, int firstColor);
	void updatePie(const QPen& pen, const Qt::BrushStyle &brushStyle, int size, int firstColor);
	void updatePieCurveData(WksheetWindow* w, const QString& yColName, int curve);
	void removePie();
	QString pieLegendText();
	QPen pieCurvePen();
	Qt::BrushStyle pieBrushStyle();
	int pieFirstColor();
	int pieSize();
	QString savePieCurveLayout();

	void plot(WksheetWindow* w,const QString& name, int style);
	void plot(WksheetWindow* w,int xcol, const QString& name, int style);	 
	void removeCurve(int index);
	void removeCurve(const QString& s);
	
	void addCurve(const QString& name);
	void updateData(WksheetWindow* w, const QString& yColName, int curve);
	void updateCurveData(WksheetWindow* w, const QString& yColName, int curve);
	 
	 int colX(int id);
	 int curves();
	 int curveDataSize(int curve);
	 double curveXValue(int curve,int point);
	 double minXValue(int curve);
	 double maxXValue(int curve);
	 double selectedXStartValue();
	 double selectedXEndValue();
	 long curveKey(int curve);
	 int curveIndex(long key);
	 void insertPlottedList(const QStringList& names);
	 QStringList plotCurvesList();
	 QStringList plotAssociations();
	 void setPlotAssociations(const QStringList& newList);

	QStringList errorBarsAssociations();
	void setErrorBarsAssociations(const QStringList& newList);

	 void print();
	 void copyImage();
	 void exportGraph();
	 void exportToEPS(const QString& fname);
	 void clearPlot();
	 void replot();
	 
	 void saveTransparentImage(const QString& fileName,const QString& fileType);
	 static void saveTransparentImage(QPixmap p,const QString& fileName,const QString& type);

	  // error bars 
	 void addErrorBars(WksheetWindow* w,const QString& name,int type, const QString& errCol);
	 void addErrorBars(WksheetWindow* w,const QString& curveName, 
										WksheetWindow* errorTable, const QString& errColName, int type);
		
	 void updateErrorBarsData(WksheetWindow* w, int curve);
	 void updateErrorBars(int curve,bool xErr,int width,int cap,
		          const QColor& c,bool plus,bool minus,bool through);
	 
	 void addErrorBars(WksheetWindow*w,bool xErrors,const QString& xColName,
						 const QString& yColName,WksheetWindow*errTable, const QString& errColName,
						 int width, int cap, const QColor& color,
						 bool through, bool minus,bool plus);
	 QStringList erorsLayout();

	 // event handlers 
	 void contextMenuEvent(QContextMenuEvent *);

	 // plot layout
	 void setPlotLimits(int axis,const QStringList& limits);
	 void setSecondaryPlotLimits(const QStringList& limits);
	 void setSecondaryPlotLimits(int axis,const QStringList& limits);
	 QStringList secondaryPlotLimits();
	 QStringList plotLimits();
	 void definePlotLimits(QStringList s);
	 QStringList xBottomAxisLimits();

	 static curveLayout initCurveLayout();
	 QArray<curveLayout> allCurvesLayout();
	 void updateCurveLayout(int curve,const curveLayout& cL);

	 gridOptions getGridOptions();
	 void setGridOptions(const gridOptions& options);

	// zoom
	 void zoomed (const QwtDoubleRect &rect);
	 void zoom(bool on);
	 bool zoomOn();
	 
	 void movedPicker(const QPoint &pos);
	 void setAutoScale();
	 void updateScale();
	
	 // saving to file 
	 QString saveToString();
	 QString saveGridOptions();
	 QString saveScale();
	 QString saveScaleTitles();
	 QString saveFonts();
	 QString saveLegend();
	 QString saveMarkers();
	 QString saveCurveLayout(int index);
	 QString saveAxesTitleColors();
	 QString saveAxesColors();
	 QString saveEnabledAxes();
	 QString saveSecondaryPlotLimits();
	 QString saveErrorBars();
	 QString saveCanvasFrame();
	 QString saveTitle();
	 QString saveAxesTitleAlignement();
	 QString saveEnabledTickLabels();
	 QString saveTicksType();
	 QString saveCurves();
	 QString saveLabelsFormat();
	 QString saveLabelsRotation();
	 QString saveAxesLabelsType();

	 // text markers 
	 void drawText(bool on);
	 bool drawTextActive();
	 
	 void insertTextMarker(LegendMarker* mrk);
	 void insertTextMarker(QStringList list);
	 void updateTextMarker(const QString& text,int angle, int bkg,const QFont& fnt,const QColor& c);
	
	 void removeMarker();
	 void cutMarker();
	 void copyMarker();
	 void pasteMarker();
	 void selectNextMarker();
	 void selectLineMarker(long markerID);
	 void selectTextMarker(long markerID);
	 void selectImageMarker(long markerID);
	 
	 MarkerType copiedMarkerType(){return selectedMarkerType;};
	 void setCopiedMarkerType(Graph::MarkerType type){selectedMarkerType=type;};
	 void setCopiedMarkerEnds(const QPoint& start, const QPoint& end);
 	 void setCopiedTextOptions(int bkg, const QString& text, const QFont& font, const QColor& color);
	 void setCopiedArrowOptions(int width, Qt::PenStyle style, const QColor& color,
													bool start, bool end, int headLength, int headAngle, bool filledHead);	
	 void setCopiedImageName(const QString& fn){auxMrkFileName=fn;};	
	 QRect copiedMarkerRect(){return QRect(auxMrkStart, auxMrkEnd);};
	 
	 //legendMarker
	 QwtArray<long> textMarkerKeys();
	 LegendMarker* textMarker(long id);
	 
	  // legend  
	 LegendMarker* legend();
	 void removeLegend();
	 void removeLegendItem(LegendMarker* mrk, int index);
	 void insertLegend(QStringList list);
	 void newLegend(const QFont& fnt, int frameStyle);
	 QSize newLegend(const QString& text);
	 bool legendOn();
	 
	 QString getLegendText();
	 void setLegendText(const QString& text);
	 
	 void setCurveLegend(const QString& text,int key);
	 QString legendText();//recreates the text from curves titles
	 
	 // line markers 
	 LineMarker* lineMarker(long id);
	  void insertLineMarker(LineMarker* mrk);
	 void insertLineMarker(QStringList list);
	 void updateLineMarker(const QColor& c,int w,Qt::PenStyle style,bool endArrow, bool startArrow);
	 QwtArray<long> lineMarkerKeys();
	 void drawLine(bool on);
	 bool drawLineActive();
	 void setArrowHeadGeometry(int length, int angle, bool filled);
	 void updateLineMarkerGeometry(const QPoint& sp,const QPoint& ep);

	 //image markers
	 ImageMarker* imageMarker(long id);
	 QwtArray<long> imageMarkerKeys();
	 void insertImageMarker(ImageMarker* mrk);
	 void insertImageMarker(const QPixmap& photo, const QString& fileName);
	 void insertImageMarker(const QStringList& options);
	 bool imageMarkerSelected();
	 void updateImageMarker(int x, int y, int width, int height);
	 
	 long selectedMarkerKey();
	 void setSelectedMarker(long mrk);
	  QwtPlotMarker* selectedMarkerPtr();
	 bool markerSelected();
	 bool arrowMarkerSelected();
	 void deselectItems();

	 // axis 
	 QValueList<int> axesType();
	 void setAxesType(const QStringList& tl);
	 QStringList axesLabelsDatasheets();

	 QStringList scalesTitles();
	 void setXAxisTitle(const QString& text);
	 void setYAxisTitle(const QString& text);
	 void setRightAxisTitle(const QString& text);
	 void setTopAxisTitle(const QString& text);
	 void setAxisTitle(int axis, const QString& text);

	 QFont axisTitleFont(int axis);
	 void setXAxisTitleFont(const QFont &fnt);
	 void setYAxisTitleFont(const QFont &fnt);
	 void setRightAxisTitleFont(const QFont &fnt);
	 void setTopAxisTitleFont(const QFont &fnt);
	 void setAxisTitleFont(int axis,const QFont &fnt);

	 void setAxisFont(int axis,const QFont &fnt);
	 QFont axisFont(int axis);
	 void initFonts(const QFont &scaleTitleFnt, const QFont &numbersFnt);
	
	 QColor axisTitleColor(int axis);
	 void setXAxisTitleColor(const QColor& c);
	 void setYAxisTitleColor(const QColor& c);
	 void setRightAxisTitleColor(const QColor& c);
	 void setTopAxisTitleColor(const QColor& c);
	 void setAxesTitleColor(QStringList l);
	 
	 int axisTitleAlignment (int axis);
	 void setXAxisTitleAlignment(int align);
	 void setYAxisTitleAlignment(int align);
	 void setTopAxisTitleAlignment(int align);
	 void setRightAxisTitleAlignment(int align);
	 void setAxesTitlesAlignment(const QStringList& align);

	 QStringList axesColors();
	 void setAxesColors(const QStringList& colors);
	 void setAxisTickLabels(int axis, int type, const QString& labelsColName, WksheetWindow *table);
	 void showAxis(int axis, int type, const QString& labelsColName, WksheetWindow *table, bool axisOn, int ticksType, 
					bool labelsOn, const QColor& c, int format, int prec, int rotation);

	 QArray<bool> enabledAxes();
	 void enableAxes(QArray<bool> axesOn);
	 void enableAxes(const QStringList& list);	

	int labelsRotation(int axis);
	void setAxisLabelRotation(int axis, int rotation);
	
	QStringList enabledTickLabels();
	void setEnabledTickLabels(const QStringList& list);
	
	QValueList<int> ticksType();
	void setTicksType(const QValueList<int>& list);
	void setTicksType(const QStringList& list); 
		
	QStringList labelsFormat();
	void setLabelsFormat(const QStringList& list);
	
	 // canvas frame 
	 void drawCanvasFrame(bool frameOn);
	 void drawCanvasFrame(const QStringList& frame);
	 void drawCanvasFrame(bool frameOn, int width, const QColor& color);
	 QColor canvasFrameColor();
	 int canvasFrameWidth();
	 bool framed();

	  // plot title 
	 QString title();
	 void setTitle(const QString& t);
	 void setTitleFont(const QFont &fnt);
	 QFont titleFont();
	 QColor titleColor();
	 void setTitleColor(const QColor &c);
	 void removeTitle();
	 bool titleSelected();
	 void selectTitle();
	 void setTitleSelected(bool on);
	 int titleAlignment();
	 void setTitleAlignment(int align);
	 void initTitle( bool on);
	 void initTitleFont( const QFont& fnt);
	
	 // tools for modifing plot data 
	 void selectPoint(const QPoint &pos);
	 void highlightPoint(bool showIt);
	 void selectCurve(const QPoint &pos);
	 int selectedCurveID();
	 
	 void showCursor(bool showIt);
	 void shiftPointCursor(bool up);
	 void shiftCurveCursor(bool up);
	 void moveBy(int dx, int dy);
	 void move(const QPoint &pos);
	 void removePoint();
	 void movePoints(bool enabled);
	 bool movePointsActivated();
	 void removePoints(bool enabled);
	 bool removePointActivated();

	 void copyCanvas(bool on);
	 void enableCursor(bool on);
	 bool enabledCursor();
	 void showPlotPicker(bool on);
	 bool pickerActivated();
	
	// data range selectors
   	 bool selectorsEnabled();
	 bool enableRangeSelectors(bool on);
	 void disableRangeSelectors();
	 void moveRangeSelector(bool up);
	 void moveRangeSelector();
	 void shiftRangeSelector(bool shift);
	 void shiftCurveSelector(bool up);
	 int selectedPoints(long curveKey);

	 //functions in analysis.cpp file
	 QString integrateCurve(long curveKey,int order,int iter,double tol,double low,double up);
	 void diffCurve(long curveKey);
	 void fft(long curveKey, const QString & FFT);
	
	 QString fitCurveLinear(long curveKey);
	 QString fitExpDecay(long curveKey);
	 QString fitExpDecay(const QString& name, double damping, double amplitude, double yOffset);
	 QString fitExpDecay2(const QString& name, double firstTime, double secondTime,
						double start, double yOffset);
	 QString fitExpDecay3(const QString& name, double firstTime, double secondTime,
						double thirdTime,double start, double yOffset);
	 QString fitExpGrowth(long curveKey);
	 QString fitGauss(long curveKey);
	 QString fitLorentz(long curveKey);
	 QString fitPolynomial(const QString&,int order, int points,
				double start, double end, bool showFormula);
	 void setFitID(int id);
	 QString fitNonlinearCurve(const QString& curve,const QString& formula,
							const QStringList& params,const QStringList& paramsInit,
							double from,double to,int points, int solver, double tolerance);
							
	void interpolation(long curveKey, const QString & Interp);

	//histograms
	void initHistogram(long curveID, const QArray<double>& Y, int it);
	void updateHistogram(WksheetWindow* w, const QString& curveName, int curve);
	void updateHistogram(WksheetWindow* w, const QString& curveName, int curve, bool automatic, 
							double binSize, double begin, double end);
	void updateBarsGap(int curve, int gapPercent);
	QString showHistogramStats(WksheetWindow* w, const QString& curveName, int curve);

	//image analyse tools
	 bool lineProfile();
	 void calculateProfile(int average, bool ok);
	 void calculateLineProfile(const QPoint& start, const QPoint& end);
	 int averageImagePixel(const QImage& image, int px, int py, int average, bool moreHorizontal);
	 void showIntensityTable();
	 
	 //user defined functions
	 void addFunctionCurve(QString& type,QStringList &formulas,QStringList &vars,QValueList<double> &ranges,QValueList<int> &points);	 
	 //when reading from file
	 void insertFunctionCurve(const QString& formula, double from, double to, int points);
	 //for versions <0.4.3
	 void insertOldFunctionCurve(const QString& formula, double from, double step, int points);

	 void createWorksheet(const QString& name);
	 void activateGraph();
	 void moveGraph(const QPoint& pos);
	 void resizeGraph(const QPoint& pos);
	 void releaseGraph();
	 void newSizeGraph();
	 void modified();

//vectXYXY curves
	void plotVectXYXY(WksheetWindow* w, const QStringList& colList);
	void updateVectorsLayout(WksheetWindow *w, int curve, int colorIndex, int width, 
								int arrowLength, int arrowAngle, bool filled, 
								const QString& xEndColName, const QString& yEndColName);
	void updateVectXYXY(WksheetWindow* w,  int curve);
						
	bool ignoreResizeEvents(){return ignoreResize;};
	void setIgnoreResizeEvents(bool ok){ignoreResize=ok;};
	void resizeMarkers ();

signals:
    void releaseGraph(Graph*);
	void newSizeGraph(Graph*);
	void moveGraph(Graph*, const QPoint& pos);
	void resizeGraph(Graph*, const QPoint& pos);
    void selectedGraph (Graph*);
	void closedGraph(QWidget*);
	void addCurveToGraph (Graph*,const QString&);
	void drawTextOff();
	void drawLineOff();
	void cursorInfo(const QString&);
	void showPlotDialog();
	void showPieDialog();
	void createTable(const QString&,int,int,const QString&);
	void updateTable(const QString&,int,const QString&);
	void clearCell(const QString&,double);

	void viewImageDialog();
	void viewTextDialog();
	void viewLineDialog();
	void viewTitleDialog();
	void modifiedGraph();
	void modifiedGraph(Graph *);
	void hiddenPlot(QWidget*);

	void showContextMenu();
	void showMarkerPopupMenu();
	
	void xAxisDblClicked();
	void yAxisDblClicked();
	
	void xAxisTitleDblClicked();		
	void yAxisTitleDblClicked();
	void rightAxisTitleDblClicked();
	void topAxisTitleDblClicked();
	
	void createTablePlot(const QString&,int,int,const QString&);
	void createIntensityTable(const QPixmap&);
	void createHistogramTable(const QString&,int,int,const QString&);
	
private:
	QStringList axesLabelsColumns;//stores columns used for axes with text tick labels
	QValueList <int> axisType;
	QValueList <int> lblFormat; //stores label format used for the axes
	QwtDiMap xCanvasMap, yCanvasMap;
	gridOptions grid;
	MarkerType selectedMarkerType;
	QwtMarker::LineStyle mrklStyle;
	QStringList scales,associations,errorBarsLayout,errorBars;
	QArray<curveLayout> layout;
	QArray<long> keys;// curves on plot keys
	QArray<long> lines; // arrows on plot keys
	QArray<long> images; // images on plot keys
	QArray<int> xCols; // x columns indexes for curves on plot
	QStringList tickLabelsOn;// store wich axes have tick labels enabled
	QPen mrkLinePen;
	QFont auxMrkFont;
	QColor auxMrkColor;
	QPoint auxMrkStart,auxMrkEnd;
	Qt::PenStyle auxMrkStyle;
	QString auxMrkFileName, auxMrkText;

	int selectedCurve, selectedPoint,startPoint,endPoint, selectedCursor, pieRay;
	int selectedCol,xCol,widthLine,cOnPlot,fitID,linesOnPlot;
	int auxMrkAngle,auxMrkBkg,auxMrkWidth, averagePixels;
	int auxArrowHeadLength, auxArrowHeadAngle;
	long  selectedMarker,legendMarkerID, startID, endID, functions;
	long mrkX,mrkY;//x=0 et y=0 line markers keys
	bool startArrowOn, endArrowOn,pickerEnabled;
	bool drawTextOn,drawLineOn,cursorEnabled;
	bool removePointsEnabled,movePointsEnabled,rangeSelectorsEnabled,piePlot;
	bool lineProfileOn; // tells if pixel line profile is asked
	bool isTitleSelected, auxFilledArrowHead, ignoreResize;
};

class PrintFilter: public QwtPlotPrintFilter
{
public:
    PrintFilter(QwtPlot *plot) 
	{
	gridMajorColor=plot->gridMajPen().color();
	gridMinorColor=plot->gridMinPen().color();		
	};

	virtual QColor color(const QColor &c, Item item, int) const
    {
        if ( !(options() & PrintCanvasBackground))
        {
            switch(item)
            {
            case MajorGrid:
                return gridMajorColor;
            case MinorGrid:
                return gridMinorColor;
            default:	
                ;
            }
        }
        return c;
    }
	
	virtual QFont font(const QFont &f, Item, int) const
    {
        QFont f2 = f;
        f2.setPointSize(int(f.pointSize() * 1.5));
        return f2;
    }

private:
	QColor gridMajorColor,gridMinorColor;
};

class Plot: public QwtPlot
{	
    Q_OBJECT

public:	
	Plot(QWidget *parent = 0, const char *name = 0);
	
	 enum TicksType{In=-1, None = 0, Out=1, Both=2};
	 
	 QValueList <int> getTicksType(){return ticksType;};
	 void setTicksType(int axis, int type);

protected:
	void printCanvas(QPainter *painter, const QRect &canvasRect,
   			 const QwtArray<QwtDiMap> &map, const QwtPlotPrintFilter &pfilter) const;
	
	virtual void drawCanvasItems (QPainter *painter, const QRect &rect,
			const QwtArray< QwtDiMap > &map, const QwtPlotPrintFilter &pfilter) const;

	void drawInwardTicks(QPainter *painter, const QRect &rect, 
							const QwtDiMap&map, int axis) const;

	void drawInwardMinorTicks(QPainter *painter, const QRect &rect, 
							const QwtDiMap &map, int axis) const;

private:
	int minTickLength, majTickLength;
	QValueList <int> ticksType;
};

#endif // GRAPH_H
